import pytest
import logging

from pdtable.table_origin import NullInputIssueTracker, InputError


def test_null_input_issue_tracker_error():
    tracker = NullInputIssueTracker()
    with pytest.raises(InputError):
        tracker.add_error("This is bad")


def test_null_input_issue_tracker_warn(caplog):
    tracker = NullInputIssueTracker()
    tracker.add_warning("This is less bad")

    # entries are logging.LogRecord instances
    # https://docs.python.org/3/library/logging.html#logrecord-attributes
    warns = [r for r in caplog.records if r.levelno >= logging.WARNING]
    assert len(warns) == 1
    assert "This is less bad" in warns[0].msg



"""Machinery to read/write Tables in an Excel workbook using openpyxl as engine."""
from itertools import chain
from os import PathLike
from typing import Union, Iterable, Sequence, Any, Dict, List, Tuple, Optional
from contextlib import closing
import openpyxl

try:
    from openpyxl.worksheet.worksheet import Worksheet as OpenpyxlWorksheet
except ImportError:
    # openpyxl < 2.6
    from openpyxl.worksheet import Worksheet as OpenpyxlWorksheet
from openpyxl.cell.cell import Cell
from openpyxl.styles import Font, PatternFill, Color, Alignment
from openpyxl.utils import get_column_letter

from pdtable import Table
from pdtable.io._represent import _represent_row_elements, _represent_col_elements

DEFAULT_SHEET_NAME = "Sheet1"
DEFAULT_STYLE_SPEC = {
    "table_name": {
        "font": {"color": "1F4E78", "bold": True,},  # hex color code
        "fill": {"color": "D9D9D9",},  # RGB color code
    },
    "destinations": {"font": {"color": "808080", "bold": True,}, "fill": {"color": "D9D9D9",},},
    "column_names": {"fill": {"color": "F2F2F2",}, "font": {"bold": True,},},
    "units": {"fill": {"color": "F2F2F2",},},
    "values": {},
}


def read_cell_rows_openpyxl(path: Union[str, PathLike]) -> Iterable[Sequence[Any]]:
    """Reads from an Excel workbook, yielding one row of cells at a time."""

    with closing(openpyxl.load_workbook(path, read_only=True, data_only=True, keep_links=False)) as wb:
        for ws in wb.worksheets:
            yield from ws.iter_rows(values_only=True)

def read_sheets(path: Union[str, PathLike]) -> Iterable[Tuple[str, Iterable[Sequence[Any]]]]:
    """Reads from an Excel workbook, yielding (sheet_name, <row iterator>)."""

    with closing(openpyxl.load_workbook(path, read_only=True, data_only=True, keep_links=False)) as wb:
        for ws in wb.worksheets:
            yield   (ws.title, ws.iter_rows(values_only=True))


def write_excel_openpyxl(tables, path, na_rep, styles, sep_lines):
    """Write tables to an Excel workbook at the specified path."""

    if not isinstance(tables, Dict):
        # For convenience, pack it in a dict
        tables = {DEFAULT_SHEET_NAME: tables}

    wb = openpyxl.Workbook()
    wb.remove(wb.active)  # Remove the one sheet that openpyxl creates by default

    for sheet_name in tables:

        tabs = tables[sheet_name]
        if isinstance(tabs, Table):
            # For convenience, pack single table in an iterable
            tabs = [tabs]

        table_dimensions = []
        ws = wb.create_sheet(title=sheet_name)
        for t in tabs:
            # Keep track of table dimensions for formatting as tuples (num rows, num cols, transposed)
            table_dimensions.append((len(t.df), len(t.df.columns), t.metadata.transposed))
            _append_table_to_openpyxl_worksheet(t, ws, sep_lines, na_rep)

        if styles:
            styles = DEFAULT_STYLE_SPEC if styles is True else styles
            _style_tables_in_worksheet(ws, table_dimensions, styles, sep_lines)

    wb.save(path)


def _append_table_to_openpyxl_worksheet(
    table: Table, ws: OpenpyxlWorksheet, sep_lines: int, na_rep: str = "-"
) -> None:
    """Write table at end of sheet, leaving sep_lines blank lines before."""
    units = table.units
    if table.metadata.transposed:
        ws.append([f"**{table.name}*"])
        ws.append([" ".join(str(x) for x in table.metadata.destinations)])
        for col in table:
            ws.append(
                [str(col.name), str(col.unit)]
                + list(_represent_col_elements(col.values, col.unit, na_rep)),
            )
    else:
        ws.append([f"**{table.name}"])
        ws.append([" ".join(str(x) for x in table.metadata.destinations)])
        ws.append(table.column_names)
        ws.append(units)
        for row in table.df.itertuples(index=False, name=None):
            # TODO: apply format string specified in ColumnMetadata
            ws.append(_represent_row_elements(row, units, na_rep))

    for _ in range(sep_lines):
        ws.append([])  # blank row marking table end


def deep_get(dictionary, keys, default=None):
    """Get value from nested dictionaries.

    Modified from: https://stackoverflow.com/a/50173148/119775

    Example:
        d = {'meta': {'status': 'OK', 'status_code': 200}}
        deep_get(d, ['meta', 'status_code'])          # => 200
        deep_get(d, ['garbage', 'status_code'])       # => None
        deep_get(d, ['meta', 'garbage'], default='-') # => '-'
    """
    if dictionary is None:
        return default
    if not keys:
        return dictionary
    return deep_get(dictionary.get(keys[0]), keys[1:], default)


def _style_cells(cells: Iterable[Cell], style: Optional[Dict]) -> None:
    if style is None:
        # Do nothing
        return

    # Font: blindly assume JSON schema matches Font.__init__() parameters (reasonable enough)
    font_args = style.get("font")
    try:
        font = Font(**font_args) if font_args else None
    except (TypeError, ValueError) as err:
        raise ValueError("Invalid font specification", font_args) from err
    # Fill: the only relevant thing is the fill color
    fill_color = deep_get(style, ["fill", "color"])
    try:
        fill = PatternFill(start_color=fill_color, fill_type="solid") if fill_color else None
    except (TypeError, ValueError) as err:
        raise ValueError("Invalid fill color", fill_color) from err
    # Alignment: blindly assume JSON schema matches Alignment.__init__() params (reasonable enough)
    alignment_args = style.get("alignment")
    try:
        alignment = Alignment(**alignment_args,) if alignment_args else None
    except (TypeError, ValueError) as err:
        raise ValueError("Invalid alignment specification", alignment_args) from err

    for cell in cells:
        # Code inspection complains that Cell attributes are read-only, but mutating them is,
        # in fact, the correct, documented way of applying styles to cells.
        # Ref: https://openpyxl.readthedocs.io/en/stable/styles.html#cell-styles
        if font is not None:
            cell.font = font  # noqa
        if fill is not None:
            cell.fill = fill  # noqa
        if alignment is not None:
            cell.alignment = alignment  # noqa


def _style_tables_in_worksheet(
    ws: OpenpyxlWorksheet,
    table_dimensions: List[Tuple[int, int, bool]],
    styles: Dict,
    sep_lines: int,
) -> None:
    num_header_rows = 2
    num_name_unit_rows = 2

    rows = [row for row in ws.iter_rows()]
    i_start = 0

    # Loop through tables
    for i, (num_rows, num_cols, transposed) in enumerate(table_dimensions):
        # Figure out on what rows this table's various parts are located
        true_num_cols = num_cols
        true_num_rows = num_rows + num_name_unit_rows
        if transposed:
            # By definition, swap understanding of rows and columns
            true_num_cols, true_num_rows = true_num_rows, true_num_cols
        table_rows = [
            r[0:true_num_cols] for r in rows[i_start : i_start + true_num_rows + num_header_rows]
        ]

        table_name_cells = table_rows[0]
        destination_cells = table_rows[1]
        if transposed:
            column_name_cells = [t[0] for t in table_rows[2:]]
            column_unit_cells = [t[1] for t in table_rows[2:]]
            value_cells = [t[2:] for t in table_rows[2:]]
        else:
            column_name_cells = table_rows[2]
            column_unit_cells = table_rows[3]
            value_cells = table_rows[4:]

        # Apply all the styles
        style_spec_names = [
            (table_name_cells, "table_name"),
            (destination_cells, "destinations"),
            (column_name_cells, "column_names"),
            (column_unit_cells, "units"),
            (chain.from_iterable(value_cells), "values"),
        ]
        for cells, style_spec_name in style_spec_names:
            try:
                _style_cells(cells, styles.get(style_spec_name))
            except ValueError as err:
                raise ValueError(f"Invalid style specification for '{style_spec_name}'") from err

        # Special default case for transposed tables: center values and units
        if transposed:
            centered = {"alignment": {"horizontal": "center"}}
            if not deep_get(styles, ["units", "alignment", "horizontal"]):
                _style_cells(column_unit_cells, centered)
            if not deep_get(styles, ["values", "alignment", "horizontal"]):
                _style_cells(chain.from_iterable(value_cells), centered)

        i_start += true_num_rows + num_header_rows + sep_lines

    # Widen columns
    max_num_cols = 0
    for rows, cols, transposed in table_dimensions:
        true_num_cols = rows if transposed else cols
        max_num_cols = max(max_num_cols, true_num_cols)
    for i_column in [get_column_letter(i + 1) for i in range(max_num_cols + 1)]:
        ws.column_dimensions[i_column].width = 20



from __future__ import annotations
from dataclasses import dataclass, field
from typing import (
    TypeVar,
    Callable,
    Iterator,
    Iterable,
)

from pdtable.proxy import Table
from pdtable.table_origin import LoadLocation

T = TypeVar("T")


@dataclass
class LocationTreeNode:
    """
    A tree structure with each node mapping to a LoadLocation

    Leaf nodes will correspond to LocationBlock objects with the corresponding
    table available as ``.table``.
    """

    location: LoadLocation
    table: None | Table = None
    parent: None | "LocationTreeNode" = None
    children: list["LocationTreeNode"] = field(default_factory=list)

    def add_child(self, child: LocationTreeNode):
        self.children.append(child)
        child.parent = self

    def visit_all(
        self, visitor: Callable[[int, "LocationTreeNode"], T], level: int = 0
    ) -> Iterator[T]:
        """
        Visitor will be called as (level, node) on self and children

        Results available as iterator.
        """
        yield visitor(level, self)
        for child in self.children:
            yield from child.visit_all(level=level + 1, visitor=visitor)

    def __str__(self) -> str:
        def str_visitor(level, node):
            if node.table is not None:
                return f"{'  '*level}**{node.table.name}"
            else:
                return (
                    f"{'  '*level}{node.location.interactive_identifier if node.location else ''}"
                )

        return "\n".join(self.visit_all(visitor=str_visitor))


def make_location_trees(tables: Iterable[Table]) -> list[LocationTreeNode]:
    """
    Return a graph representation of the origins for given tables

    Graph is a collection of trees:
        B LocationFile
          L LocationBlock # table
          B LocationBlock # include
            B LocationFile
                    ....
                L LocationBlock  # table
        B LocationFolder
        B LocationFile
            L LocationBlock

    The implementation relies on ``load_identifier`` to be unique for a LocationFile object.

    Return:
        List of root origins
    """
    # Implementation note:
    # To extend to non-compliant readers that do not implement unique load identifiers,
    # a check for object identify could be added.

    # Create leaf nodes and immediate parents, track by load_identifier
    # Since load_identifier for LocationBlock just refers to parent file,
    # individual tables are not registered
    buf: dict[str, LocationTreeNode] = {}

    def register_node(location: LoadLocation, child: LocationTreeNode):
        """
        Add location and ancestors to buf by load_identifier

        Should not be called for individual LocationBlock
        """
        if location.load_identifier in buf:
            buf[location.load_identifier].add_child(child)
            return
        new_node = LocationTreeNode(location=location)
        new_node.add_child(child)
        buf[location.load_identifier] = new_node
        source = location.load_specification.source
        if source is not None:
            register_node(source, child=new_node)

    for t in tables:
        if t.metadata.origin is None:
            raise ValueError("Table object without origin not supported for `make_origin_trees`", t)
        location = t.metadata.origin.input_location
        if location is None:
            if t.metadata.origin.parents:
                raise NotImplementedError("Support for non-loaded tables not implemented")
            else:
                raise ValueError("Missing input_location for table", t)
        leaf = LocationTreeNode(location=location, table=t)
        register_node(location.file, child=leaf)

    # return nodes without parent as roots:
    return [v for v in buf.values() if v.parent is None]



import json
import os
from pathlib import Path
from textwrap import dedent

import pandas as pd
import pytest
from pytest import raises

from pdtable import ParseFixer, BlockType
from pdtable import read_csv
from pdtable.table_origin import InputError
from pdtable.io import table_to_json_data
from pdtable.io.parsers import parse_blocks
from pdtable.io.parsers.blocks import make_table


class custom_test_fixer(ParseFixer):
    def __init__(self):
        super().__init__()
        self.stop_on_errors = False
        self._called_from_test = True


def input_dir() -> Path:
    return Path(__file__).parent / "input/with_errors"


def test_displays_all_error_messages():
    """By default, ParseFixer stops on errors and outputs a message
    listing all encountered errors."""
    expected_error_msg = dedent(
        """\
        Stopped parsing after 2 errors in table 'farm_cols1' with messages:
        Duplicate column 'flt' at position 4 in table 'farm_cols1'.
        Duplicate column 'flt' at position 5 in table 'farm_cols1'."""
    )
    with raises(InputError) as input_error:
        blocks = list(read_csv(input_dir() / "cols1.csv"))

    msg = input_error.value.args[0].issue  # Avoid repr escaping
    assert expected_error_msg == str(msg)

def test_columns_duplicate():
    """
       Verify that default ParseFixer corrects duplicate column names

    """
    tab = None
    with open(input_dir() / "cols1.csv", "r") as fh:
        g = read_csv(fh, fixer=custom_test_fixer)
        for tp, tt in g:
            if True:
                if tp == BlockType.TABLE:
                    tab = tt
                    break
    assert tab is not None
    assert tab.df["flt_fixed_001"] is not None
    assert tab.df["flt_fixed_001"][6] == 7.6
    assert tab.df["flt"][0] == 3.0


def test_custom_fixer():
    """ Test custom ParseFixer
        Verify that read_csv uses custom ParseFixer
    """

    class fix_pi(ParseFixer):
        def __init__(self):
            super().__init__()

        # augment existing method, simple fix float
        def fix_illegal_cell_value(self, vtype, value):
            if vtype == "float":
                return 22.0 / 7.0
            else:
                fix_value = ParseFixer.fix_illegal_cell_value(self, vtype, value)
                return fix_value

    fix = fix_pi()
    fix.stop_on_errors = False
    fix._called_from_test = True

    with open(input_dir() / "types3.csv", "r") as fh:
        g = read_csv(fh, to="jsondata", fixer=fix)
        for tp, tt in g:
            if tp == BlockType.TABLE:
                assert tt["columns"]["num"]["values"][2] == 22.0 / 7.0
                assert tt["columns"]["flt"]["values"][0] == 22.0 / 7.0
                assert tt["columns"]["flt"]["values"][0] == 22.0 / 7.0
                assert tt["columns"]["flt2"]["values"][2] == 22.0 / 7.0

    with pytest.raises(InputError):
        # test read_csv w. class (not instance) of fixer
        # class has default stop_on_errors = True
        with open(input_dir() / "types3.csv", "r") as fh:
            g = read_csv(fh, to="jsondata", fixer=fix_pi)
            for tp, tt in g:
                pass


def test_FAT():
    """ Factory Acceptance Test

        Verify that we are able to read all files in ./input
        Using default ParseFixer
    """
    all_files = 0
    ignore_files = ["auto_fixed.py", "__init__.py", "all.json"]
    for fn in os.listdir(input_dir()):
        path = input_dir() / fn
        if not os.path.isfile(path):
            continue
        if fn in ignore_files:
            continue
        all_files += 1

    # load targets
    with open(input_dir() / "all.json") as f:
        all_json = json.load(f)

    for fn in os.listdir(input_dir()):
        path = input_dir() / fn
        if not os.path.isfile(path):
            continue
        if fn in ignore_files:
            continue

        with open(input_dir() / fn, "r") as fh:
            g = read_csv(fh, origin=f'"{fn}"', to="jsondata", fixer=custom_test_fixer)
            count = 0
            for tp, tt in g:
                if tp == BlockType.TABLE:
                    count += 1
                    if fn != "all.csv":
                        assert tt == all_json[fn]

            if fn == "all.csv":
                assert count == all_files - 1
            else:
                assert count == 1


def test_stop_on_errors():
    """ Unit test cusrom ParseFixer.stop_on_errors
    """
    # fmt: off
    table_lines = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ 1     , 2     , 3     , 3.14  ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ "NaN" , "nan" , "Nine", "Ten" ],
        [ 1     , 2     , 3     , 4     ],
    ]
    # fmt: on

    fix = ParseFixer()
    fix.stop_on_errors = True
    fix._dbg = False  # ignore during test
    pi = 0
    with pytest.raises(InputError):
        for typ, tab in parse_blocks(table_lines, fixer=fix, to="pdtable"):
            if typ != BlockType.TABLE:
                continue
            assert tab.df["a4"][0] == 3.14
            pi += 1

    with pytest.raises(InputError):
        for typ, tab in parse_blocks(table_lines, fixer=fix, to="jsondata"):
            if typ != BlockType.TABLE:
                continue
            assert tab["columns"]["a4"]["values"][0] == 3.14
            pi += 1

    # cellgrid does not parse values, i.e. no ValueError
    for typ, tab in parse_blocks(table_lines, fixer=fix, to="cellgrid"):
        if typ != BlockType.TABLE:
            continue
        if tab[0][0] == "**tab_ok":
            assert tab[4][3] == 3.14
            pi += 1
        if tab[0][0] == "**tab_errors":
            assert tab[4][3] == "Ten"

    assert pi == 3  # 😉


def test_stop_on_errors_default_fixer():
    """ Unit test ParseFixer: raise ValueError on empty float and empty onoff
    """
    # fmt: off
    table_lines_flt = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ 1     , 2     , 3     , 3.14  ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ 1     , 2     , 3     , ""    ],
    ]
    # fmt: on

    g = parse_blocks(table_lines_flt, filter=lambda ty, tn: ty == BlockType.TABLE)
    typ, tab = next(g)
    assert tab.name == "tab_ok"
    assert tab.df["a4"][0] == 3.14

    with pytest.raises(InputError):
        typ, tab = next(g)
        print(f"-oOo-: {typ} {tab}")

    # fmt: off
    table_lines_onoff = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"],
        [ "-"   , "onoff"],
        [ 3.14  , 1      ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2" ],
        [ "-"   , "onoff" ],
        [ 1     ,  None   ],
    ]
    # fmt: on
    g = parse_blocks(table_lines_onoff, filter=lambda ty, tn: ty == BlockType.TABLE)
    typ, tab = next(g)
    assert tab.name == "tab_ok"
    assert tab.df["a1"][0] == 3.14

    with pytest.raises(InputError):
        typ, tab = next(g)
        print(f"-oOo-: {typ} {tab}")

    # fmt: off
    table_lines_datetime = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"],
        [ "-"   , "datetime"   ],
        [ 14    , pd.to_datetime("2020-08-11")],
        [ 3.14  , "2020-08-12" ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2" ],
        [ "-"   , "datetime" ],
        [ 14    , pd.to_datetime("2020-08-13")],
        [ 1     ,  None   ],
    ]
    # fmt: on
    g = parse_blocks(table_lines_datetime, filter=lambda ty, tn: ty == BlockType.TABLE)
    typ, tab = next(g)
    assert tab.name == "tab_ok"
    assert tab.df["a1"][1] == 3.14
    assert tab.df["a2"][1] == pd.to_datetime("2020-08-12")

    with pytest.raises(InputError):
        typ, tab = next(g)
        print(f"-oOo-: {typ} {tab}")


def test_converter():
    """ Unit test
        Verify that misc. float input are handled consistently
    """
    # fmt: off
    table_lines = [
        ["**flt_errors"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ "NaN" , "nan" , "Nine", "Ten" ],
        [ 1     , 2     , 3     , 3.14  ],
    ]
    # fmt: on
    cf = custom_test_fixer()
    pandas_pdtab = make_table(table_lines, fixer=cf)
    js_obj = table_to_json_data(pandas_pdtab)
    assert js_obj["columns"]["a3"]["values"][0] is None
    assert js_obj["columns"]["a4"]["values"][1] == 3.14

    assert cf.fixes == 2  # Nine and Ten



"""
The ``io.load``-module is responsible for reading input sets

The ``io.load`` module is implemented as a layer on top of the format-specific single-file readers in 
the ``io``-module. The added functionality include:

  - Support of `***include` directives
  - Support for reading entire directories
  - Support for tracking input origin
  - Support for multiple file types identified by extension
  - Support for multiple input protocols (records systems, blobs, etc)
  - Future support for multithreaded readers

The ``io.load``-module relies heavily on the features of the ``table_origin`` module, but there is no inverse
dependency, i.e. the ``table-origin`` functionality can be used from alternative loader systems without any 
reference to the ``io.load``-module.

Example of loading all files of supported formats in a given folder::

    inputs = load_files(['/'], root_folder=root_folder, csv_sep=';')
    bundle =  TableBundle(inputs)

The Table objects returned by load will have an `origin` attribute describing their load herritage.
An example of how this can be used is the `make_location_trees`-function, which builds a tree-representation of the load process::

    location_trees = make_location_trees(iter(bundle))
    print('\n'.join(str(n) for n in location_trees))

The functionality of the ``load``-module is implemented in a composable form to allow it to be integrated with 
project-specific storage systems. In the simplest form, this can be done by passing loaders for additional protocols
to ``load_files``. More involved customizations can be obtained by manually composing the master loader instance. 
The implementation for ``load_files`` uses a composition of ``ProtocolLoader``, ``FileSystemLoader`` and ``IncludeLoader``.
"""
# flake8: noqa

from ._protocol import (
    LoadOrchestrator,
    Loader,
    LoadError,
)
from ._loaders import (
    FileReader,
    ProtocolLoader,
    FileSystemLoader,
    IncludeLoader,
)
from ._orchestrators import (
    load_files,
)
from ._tree import (
    LocationTreeNode,
    make_location_trees,
)



from __future__ import annotations
from typing import Iterable
import re
from pathlib import Path

from pdtable.store import BlockIterator
from pdtable.table_origin import (
    NullInputIssueTracker,
    InputIssueTracker,
    InputError,
    LoadItem,
)
from ._protocol import (
    Loader,
)
from ._loaders import make_loader, FileReader


def queued_load(roots: list[LoadItem], loader: Loader, issue_tracker: InputIssueTracker = None):
    """
    Load `LoadItem`-objects listed in `roots`, together with any items added by the loader due
    to ``***include`` directives or similar

    This loader is single threaded. For higher performance, a multi-threaded loader should be used.

    This loader does not check for include-loops.
    """

    class Orchestrator:
        def __init__(self, roots, issue_tracker):
            self.load_items = roots
            self.issue_tracker = issue_tracker

        def add_load_item(self, item):
            self.load_items.append(item)

    orch = Orchestrator(
        roots, issue_tracker if issue_tracker is not None else NullInputIssueTracker()
    )
    visited: set[str] = set()
    while orch.load_items:
        load_proxy = loader.resolve(orch.load_items.pop(), orch)
        # check for loops/duplicates
        load_identifier = load_proxy.load_location.load_identifier
        if load_identifier in visited:
            orch.issue_tracker.add_error(
                "Load location included multiple times (this may be due to an include loop)",
                load_location=load_proxy.load_location)
            continue
        visited.add(load_identifier)

        yield from load_proxy.read(orch)


def load_files(
    roots: Iterable[str | Path] = None,
    *,
    issue_tracker: None | InputIssueTracker = None,
    # below inputs are forwarded to make_reader -- only included for easy docs
    csv_sep: None | str = None,
    sheet_name_pattern: re.Pattern = None,
    file_reader: FileReader = None,
    root_folder: None | str | Path = None,
    file_name_pattern: re.Pattern = None,
    file_name_start_pattern: str = None,
    additional_protocol_loaders: dict[str, Loader] = None,
    allow_include: bool = True,
    **kwargs,
) -> BlockIterator:
    """
    Load a complete startable inputset

    Example: load all files matching `input_*`, `setup_*` in folder `foo`::

        load_files(root_folder="foo",
                   csv_sep=';', file_name_start_pattern="^(input|setup)_")

    This function is a thin wrapper around the current best-practice loader
    and the backing implementation will be updated when best practice changes.

    `load_files` uses the ``FileSystemReader`` to resolve paths, which means that you must
    pass absolute filenames. See docs for ``FileSystemReader`` for details.

    Args:
        roots: The root load items.
            If ``root_folder`` is specified, contents must be valid root load specifiers
            which cannot be relative file names. Default value is ``["/"]``, indicating
            that the root folder is the only root load item.
            If ``root_folder`` is not specified, file-protocol roots must be provided as
            absolute paths.
        issue_tracker: Optional; Custom `InputIssuesTracker` instance to use.

    Any additional keyword arguments are forwarded to `make_loader` (see docs there).
    """
    loader = make_loader(
        csv_sep=csv_sep,
        sheet_name_pattern=sheet_name_pattern,
        file_reader=file_reader,
        root_folder=root_folder,
        file_name_pattern=file_name_pattern,
        file_name_start_pattern=file_name_start_pattern,
        additional_protocol_loaders=additional_protocol_loaders,
        allow_include=allow_include,
        **kwargs,
    )
    if roots is None and root_folder is not None:
        roots = ["/"]
    yield from queued_load(
        roots=[LoadItem(str(f), source=None) for f in roots],
        loader=loader,
        issue_tracker=issue_tracker,
    )



from __future__ import annotations
from pathlib import Path
from dataclasses import dataclass
import logging
import re
from typing import (
    NamedTuple,
)

from pdtable.store import BlockIterator, BlockType
from pdtable.table_origin import (
    LoadItem,
    LoadLocation,
    FilesystemLocationFile,
    LocationFile,
    interactive_open_uri,
)

# from ..reader import FileReader
from ..csv import read_csv
from ..excel import read_excel
from ._protocol import (
    LoadProxy,
    Loader,
    Reader,
    LoadError,
    LoadOrchestrator,
)


logger = logging.getLogger(__name__)


class FileReader:
    """
    `FileReader` is a wrapper that integrates the single-file loaders with the `load`-module
    """

    csv_sep: None | str
    sheet_name_pattern: re.Pattern

    def __init__(self, sheet_name_pattern: re.Pattern, csv_sep: None | str = None):
        self.csv_sep = csv_sep
        self.sheet_name_pattern = sheet_name_pattern

    @property
    def supported_extensions(self) -> list[str]:
        return ["csv", "xlsx"]

    @property
    def supported_filename_pattern(self) -> re.Pattern:
        """
        A regex pattern matching supported file extensions

        Example value: ``re.compile(r".*\\.(xlsx|csv)$")``
        """
        return re.compile(r".*\.({})$".format("|".join(self.supported_extensions)), re.IGNORECASE)

    def read(self, location_file: LocationFile, orchestrator: LoadOrchestrator) -> BlockIterator:
        path = location_file.get_local_path()
        ext = path.suffix.lower()
        if ext == ".csv":
            location_sheet = location_file.make_location_sheet()
            yield from read_csv(
                path,
                sep=self.csv_sep,
                location_sheet=location_sheet,
                issue_tracker=orchestrator.issue_tracker,
            )
        elif ext == ".xlsx":
            yield from read_excel(
                path,
                sheet_name_pattern=self.sheet_name_pattern,
                location_file=location_file,
                issue_tracker=orchestrator.issue_tracker,
            )
        else:
            raise ValueError(f"Unsupported file extension: {ext}")


class LocationFolder(NamedTuple):
    """
    An implementation of the LoadLocation interface to describe a folder
    """

    local_folder_path: Path
    load_specification: LoadItem
    root_folder: None | Path = None

    @property
    def load_identifier(self) -> str:
        return str(self.local_folder_path)

    @property
    def interactive_identifier(self) -> str:
        if self.root_folder is None:
            return self.load_identifier
        rel_path = self.local_folder_path.relative_to(self.root_folder)
        if rel_path == Path("."):
            return f"<root_folder: {self.root_folder}>"
        else:
            return str(rel_path)

    def interactive_uri(self, read_only=False) -> str:
        return self.local_folder_path.as_uri()

    def interactive_open(self):
        interactive_open_uri(self.interactive_uri())

    @classmethod
    def make_location_folder(
        cls,
        local_folder_path: Path,
        load_specification: LoadItem = None,
        root_folder: None | Path = None,
    ) -> "LocationFolder":
        if load_specification is None:
            load_specification = LoadItem(str(local_folder_path), source=None)
        return cls(
            local_folder_path=local_folder_path,
            load_specification=load_specification,
            root_folder=root_folder,
        )


_LEADING_SLASH = re.compile(r"/|\\")


@dataclass(frozen=True)
class FolderReader:
    file_name_pattern: re.Pattern

    def read(self, location: LocationFolder, orchestrator: LoadOrchestrator) -> BlockIterator:
        for p in location.local_folder_path.iterdir():
            if not self.file_name_pattern.match(p.name):
                continue
            logger.debug(f"Including file '{p}' from '{location}'")
            orchestrator.add_load_item(LoadItem(specification=p.name, source=location))
        yield from ()  # ensure this is an iterator


@dataclass
class FileSystemLoader:
    """
    A loader protocol for local filesystem loads

    Paths are resolved as follows:
    - relative paths are resolved relative to the file/folder they originate in
    - absolute paths are resolved relative to the root folder

    ``file_name_pattern``
    ---------------------
    When a directory is a directory, files matching this pattern will be loaded.

    ``root_folder``
    ---------------
    If a ``root_folder`` is specified, all path specifications with a leading ``/`` or ``\\`` are
    resolved relative to the root folder, and all paths must be inside the root folder.
    This implies absolute paths will be handled differently depending on whether the root folder is
    defined or not, and so the decision on whether or not to use root folder in a project should
    be taken with care.

    ``ignore_protocol``
    -------------------
    When used together with `ProtocolLoader`, specification strings may include protocols.
    The ``ignore_protocol`` specifies a protocol value that should be ignored.
    All other protocol values in specification will assumed to be part of the filesystem path,
    likely leading to errors.
    Value of ``ignore_protcol`` must be lower case and include trailing ``:``.
    Default value is ``file:``.
    """

    file_reader: FileReader
    folder_reader: FolderReader
    root_folder: None | Path = None
    ignore_protocol: str = "file:"

    def _resolve_load_item_path(self, load_item: LoadItem) -> Path:
        spec = load_item.specification

        if self.ignore_protocol and spec.lower().startswith(self.ignore_protocol):
            spec = spec[len(self.ignore_protocol) :]
        resolved: Path = Path(spec)

        leading_slash = _LEADING_SLASH.match(spec) is not None
        if leading_slash:
            if self.root_folder is None:
                if not resolved.is_absolute():
                    raise LoadError(
                        "Include with leading slash must be an absolute path when root_folder "
                        "not defined"
                    )
            else:
                resolved = self.root_folder / spec[1:]
        else:
            if load_item.source is None or load_item.source.local_folder_path is None:
                if not resolved.is_absolute():
                    raise LoadError("Cannot load location relative to source with no local folder path")
            else:
                resolved = load_item.source.local_folder_path / spec
        resolved = resolved.resolve()

        if self.root_folder is not None:
            # Check that resolved folder is inside root
            try:
                resolved.relative_to(self.root_folder)
            except ValueError:
                raise LoadError(
                    f"Load item {resolved} is outside load root folder: {self.root_folder}"
                )
        return resolved

    def resolve(self, load_item: LoadItem, orchestrator: LoadOrchestrator) -> LoadProxy:
        try:
            full_path = self._resolve_load_item_path(load_item)
        except LoadError as e:
            # tracker may raise new exception if so configured, but this error cannot be recovered
            orchestrator.issue_tracker.add_error(e, load_item=load_item)
            raise e

        if full_path.is_dir():
            load_location = LocationFolder(
                local_folder_path=full_path,
                load_specification=load_item,
                root_folder=self.root_folder,
            )
            reader = self.folder_reader
        else:
            load_location = FilesystemLocationFile(
                local_path=full_path, load_specification=load_item, root_folder=self.root_folder
            )
            reader = self.file_reader

        return LoadProxy(load_location=load_location, reader=reader)


@dataclass(frozen=True)
class ProtocolLoader(Loader):
    """
    A composable Loader implementation that allows multi-protocol support

    ``default_protocol``
    -------------------
    If the ``specification`` in a ``LoadItem`` instance does not include a known protocol,
    the protocol is assumed to be ``default_protocol``.
    """

    protocol_handlers: dict[str, Loader]
    default_protocol: str = "file"

    def resolve(self, load_item: LoadItem, orchestrator: LoadOrchestrator) -> LoadProxy:
        # Do not try to parse protocol, as this may cause issues on windows where file paths may
        # look like protocols (e.g. C:...)
        spec = load_item.specification.lower()
        handler = next(
            (h for p, h in self.protocol_handlers.items() if spec.startswith(p + ":")),
            self.protocol_handlers[self.default_protocol],
        )
        return handler.resolve(load_item=load_item, orchestrator=orchestrator)


@dataclass(frozen=True)
class IncludeReader(Reader):
    reader: Reader

    def read(self, load_location: LoadLocation, orchestrator: LoadOrchestrator) -> BlockIterator:
        for block_type, value in self.reader.read(load_location, orchestrator):
            if block_type == BlockType.DIRECTIVE and value.name == "include":
                source = value.origin.input_location
                for line in value.lines:
                    logger.debug(f"Adding include '{line}' from {source.interactive_identifier}")
                    orchestrator.add_load_item(LoadItem(specification=line, source=source))
            else:
                yield block_type, value


@dataclass(frozen=True)
class IncludeLoader(Loader):
    """
    A composable Loader implementation that implements an ``include``-directive

    Each row in an include directive correspond to a load item.
    The ``FileSystemLoader`` will resolve relative file paths relative to the
    folder containing the file with the include-directive.
    """

    loader: Loader

    def resolve(self, load_item: LoadItem, orchestrator: LoadOrchestrator) -> LoadProxy:
        proxy = self.loader.resolve(load_item, orchestrator)
        return LoadProxy(reader=IncludeReader(proxy.reader), load_location=proxy.load_location)


def make_loader(
    *,
    csv_sep: None | str = None,
    sheet_name_pattern: re.Pattern = None,
    file_reader: FileReader = None,
    root_folder: None | str | Path = None,
    file_name_pattern: re.Pattern = None,
    file_name_start_pattern: str = None,
    additional_protocol_loaders: dict[str, Loader] = None,
    allow_include: bool = True,
) -> Loader:
    """
    Make a composite loader

    args:
        csv_sep:
            Optional; Separator for csv files. Defaults to pdtable.CSV_SEP
        sheet_name_pattern:
            Optional; Compiled regexp to match sheet names against when reading a workbook
            directory. Defaults to `.*`. Typical values include `^(input|setup)`
        file_reader:
            Optional; If not supplied, file_reader is instantiated based on `csv_sep` and
            `sheet_name_pattern`. These arguments cannot be provided together with ``file_reader``.
            Reasons to include external reader could include low-level filtering to improve
            performance.
        root_folder:
            Optional; Root folder for resolving absolute imports. If defined, no loads outside
            root folder are allowed. See also `FileSystemLoader`.
        file_name_start_pattern:
            Optional; If provided ``file_name_pattern`` is computed as ``file_name_start_pattern``
            combined with the extension pattern defined by the file reader.
            If not provided a negative lookahead for excel temp files is used `(?!~\\$)`
            Example value `"^(input|setup)_"`.
        file_name_pattern:
            Compiled regexp to match filenames against when reading a whole directory.
            Defaults to extension pattern defined by the file reader.
        allow_include:
            Optional; Default is True; If True, honor the ``include`` directive.
        additional_protocol_loaders:
            Optional; A dict mapping protocol names (lower case, no colon) to ``Loader``-objects
            to handle that protocol. Will be supplemented with a default file protocol.
    """

    # file reader
    if file_reader is None:
        file_reader = FileReader(csv_sep=csv_sep, sheet_name_pattern=sheet_name_pattern)
    elif csv_sep is not None or sheet_name_pattern is not None:
        raise ValueError("csv_sep and sheet_name_pattern cannot be used with file_reader")

    # folder reader
    if file_name_pattern is None:
        if file_name_start_pattern is None:
            # negative lookahead to avoid excel tempfiles
            file_name_start_pattern = r"(?!~\$)"
        sfp = file_reader.supported_filename_pattern
        file_name_pattern = re.compile(file_name_start_pattern + sfp.pattern, sfp.flags)
    elif file_name_start_pattern is not None:
        raise ValueError("file_name_start_pattern cannot be used with file_name_pattern")
    folder_reader = FolderReader(file_name_pattern=file_name_pattern)

    # filesystem loader
    if root_folder is not None:
        root_folder = Path(root_folder)
    file_loader = FileSystemLoader(
        file_reader=file_reader, folder_reader=folder_reader, root_folder=root_folder
    )

    # protocols
    loader = file_loader
    if additional_protocol_loaders is not None:
        loader = ProtocolLoader(
            handlers={
                kv
                for ll in [[("file", file_reader)], additional_protocol_loaders.items()]
                for kv in ll
            },
            default_protocol="file",
        )

    # include directive
    if allow_include:
        loader = IncludeLoader(loader)

    return loader



from __future__ import annotations
from abc import abstractmethod
import logging
import typing

from typing_extensions import Protocol  # Protocol is not available in python 3.7

from pdtable.store import BlockIterator
from pdtable.table_origin import (
    InputIssueTracker,
    LoadItem,
    LoadLocation,
)

logger = logging.getLogger(__name__)


class LoadError(Exception):
    pass


class LoadOrchestrator(Protocol):
    """
    The load orchestrator provides state during the load process

    In contrast to ``Loader`` and ``Reader`` instances, which should be
    reusable across load operations, the orchestrator should only be used
    once.
    """
    @abstractmethod
    def add_load_item(self, w: LoadItem):
        pass

    @property
    @abstractmethod
    def issue_tracker(self) -> InputIssueTracker:
        pass


class Reader(Protocol):
    @abstractmethod
    def read(self, load_location: LoadLocation, orchestrator: LoadOrchestrator) -> BlockIterator:
        pass


# The ``Reader`` interface could in principle be replaced with a callable type, i.e.::
#     Reader = typing.Callable[[LoadLocation, LoadOrchestrator], BlockIterator]
# This approach was decided against because most readers turned out to need internal state
# which is easier to inspect with the explicit object form rather than a partially evaluated
# callable.
# Further, it is simple to adapt a callable to the object-model, see ``CallableReader``.


class CallableReader(typing.NamedTuple):
    """
    Adapter to wrap a callable reader for the ``Reader``-interface

    Example use::
        reader = CallableReader(my_reader_function)
    """

    read: typing.Callable[[LoadLocation, LoadOrchestrator], BlockIterator]


# class LoadProxy(Protocol):
#     @property
#     @abstractmethod
#     def reader(self) -> Reader:
#         pass


#     @property
#     @abstractmethod
#     def load_location(self) -> LoadLocation:
#         pass


#     @abstractmethod
#     def read(self, orchestrator: LoadOrchestrator) -> BlockIterator:
#         pass


class LoadProxy(typing.NamedTuple):
    load_location: LoadLocation
    reader: Reader

    def read(self, orchestrator: LoadOrchestrator) -> BlockIterator:
        yield from self.reader.read(self.load_location, orchestrator)


class Loader(Protocol):
    """
    The loader is responsible for resolving the load specification into a block iterator

    The resolution result is returned as a ``LoadProxy`` instance which holds the resolved
    ``LoadLocation`` together with a ``Reader`` instance compatible with the location.
    """

    @abstractmethod
    def resolve(self, load_item: LoadItem, orchestrator: LoadOrchestrator) -> LoadProxy:
        pass

    def load(self, load_item: LoadItem, orchestrator: LoadOrchestrator) -> BlockIterator:
        yield from self.resolve(load_item, orchestrator).read(orchestrator)



import datetime

import pytest
import pandas as pd
import openpyxl

try:
    from openpyxl.worksheet.worksheet import Worksheet as OpenpyxlWorksheet
except ImportError:
    # openpyxl < 2.6
    from openpyxl.worksheet import Worksheet as OpenpyxlWorksheet  # noqa: F401

from pdtable import Table
from pdtable.io.excel import write_excel
from pdtable.io._excel_openpyxl import _append_table_to_openpyxl_worksheet, deep_get


def test_deep_get():
    d = {'meta': {'status': 'OK', 'status_code': 200}}
    assert deep_get(d, ['meta', 'status_code']) == 200
    assert deep_get(d, ['garbage', 'status_code']) is None
    assert deep_get(d, ['meta', 'garbage'], default='-') == '-'


def test__append_table_to_openpyxl_worksheet():
    # Make a table with content of various units
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    wb = openpyxl.Workbook()
    ws = wb.active

    # Act
    _append_table_to_openpyxl_worksheet(t, ws, sep_lines=1)

    # Assert worksheet looks as expected:
    # table header by row
    assert ws["A1"].value == "**foo"
    assert ws["A2"].value == "all"
    assert [ws.cell(3, c).value for c in range(1, 5)] == ["place", "distance", "ETA", "is_hot"]
    assert [ws.cell(4, c).value for c in range(1, 5)] == ["text", "km", "datetime", "onoff"]

    # table data by column
    assert [ws.cell(r, 1).value for r in range(5, 9)] == ["home", "work", "beach", "wonderland"]
    assert [ws.cell(r, 2).value for r in range(5, 9)] == [0, 1, 2, "-"]
    assert [ws.cell(r, 3).value for r in range(5, 9)] == list(
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00"])
    ) + ["-"]
    assert [ws.cell(r, 4).value for r in range(5, 9)] == [1, 0, 1, 0]


def test_write_excel(tmp_path):
    # Make a couple of tables
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    # This one is transposed
    t2 = Table(name="bar")
    t2.add_column("digit", [1, 6, 42], "-")
    t2.add_column("spelling", ["one", "six", "forty-two"], "text")
    t2.metadata.transposed = True

    # Write tables to workbook, save, and re-load
    out_path = tmp_path / "foo.xlsx"
    write_excel([t, t2], out_path)
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # First table is written as expected
    # - table header by row
    assert ws["A1"].value == "**foo"
    assert ws["A2"].value == "all"
    assert [ws.cell(3, c).value for c in range(1, 5)] == ["place", "distance", "ETA", "is_hot"]
    assert [ws.cell(4, c).value for c in range(1, 5)] == ["text", "km", "datetime", "onoff"]

    # - table data by column
    assert [ws.cell(r, 1).value for r in range(5, 9)] == ["home", "work", "beach", "wonderland"]
    assert [ws.cell(r, 2).value for r in range(5, 9)] == [0, 1, 2, "-"]
    for r, d in zip(
        range(5, 8), pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00"])
    ):
        # workaround openpyxl bug: https://foss.heptapod.net/openpyxl/openpyxl/-/issues/1493
        # openpyxl adds a spurious microsecond to some datetimes.
        assert abs(ws.cell(r, 3).value - d) <= datetime.timedelta(microseconds=1)
    assert ws.cell(8, 3).value == "-"
    assert [ws.cell(r, 4).value for r in range(5, 9)] == [1, 0, 1, 0]

    # Second table is there as well
    assert ws["A10"].value == "**bar*"
    assert ws["A11"].value == "all"
    # column headers (transposed)
    assert [ws.cell(r, 1).value for r in range(12, 14)] == ["digit", "spelling"]
    assert [ws.cell(r, 2).value for r in range(12, 14)] == ["-", "text"]
    # column values (transposed)
    assert [ws.cell(12, c).value for c in range(3, 6)] == [1, 6, 42]
    assert [ws.cell(13, c).value for c in range(3, 6)] == ["one", "six", "forty-two"]

    # Teardown
    out_path.unlink()


def test_write_excel__multiple_sheets(tmp_path):
    """write_excel() can write tables to multiple sheets in a workbook"""

    # Make a couple of tables
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    t2 = Table(name="bar")
    t2.add_column("digit", [1, 6, 42], "-")
    t2.add_column("spelling", ["one", "six", "forty-two"], "text")

    # Write tables to workbook, save, and re-load
    out_path = tmp_path / "foo.xlsx"
    write_excel({"sheet_one": [t, t2], "sheet_two": t2}, out_path)
    wb = openpyxl.load_workbook(out_path)

    # Workbook has the expected sheets
    assert len(wb.worksheets) == 2
    assert wb.sheetnames == ["sheet_one", "sheet_two"]
    # First sheet contains the expected tables
    assert wb.worksheets[0]["A1"].value == "**foo"
    assert wb.worksheets[0]["A10"].value == "**bar"
    # Second sheet contains the expected tables
    assert wb.worksheets[1]["A1"].value == "**bar"
    # Table details are tested elsewhere.

    # Teardown
    out_path.unlink()


def test_write_excel__style(tmp_path):
    # Make a couple of tables
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    # This one is transposed
    t2 = Table(name="bar")
    t2.add_column("digit", [1, 6, 42], "-")
    t2.add_column("spelling", ["one", "six", "forty-two"], "text")
    t2.metadata.transposed = True

    # This one is also transposed
    t3 = Table(name="bas")
    t3.add_column("digit", [1, 6, 42], "-")
    t3.add_column("spelling", ["one", "six", "forty-two"], "text")
    t3.metadata.transposed = True

    # Write tables to workbook, save, and re-load
    out_path = tmp_path / "foo.xlsx"
    write_excel([t, t2, t3], out_path, styles=True)
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # First table is written as expected
    # - table header by row
    assert ws["A1"].value == "**foo"
    assert ws["A2"].value == "all"
    assert [ws.cell(3, c).value for c in range(1, 5)] == ["place", "distance", "ETA", "is_hot"]
    assert [ws.cell(4, c).value for c in range(1, 5)] == ["text", "km", "datetime", "onoff"]

    # Check table formatting
    assert ws["A1"].fill.fill_type == "solid"
    assert ws["A1"].fill.start_color.value == "00D9D9D9"
    assert ws["A1"].font.color.value == "001F4E78"
    assert ws["A1"].font.bold is True

    assert ws["A2"].fill.fill_type == "solid"
    assert ws["A2"].fill.start_color.value == "00D9D9D9"
    assert ws["A2"].font.color.value == "00808080"
    assert ws["A2"].font.bold is True

    assert [ws.cell(3, c).fill.fill_type for c in range(1, 5)] == ["solid"] * 4
    assert [ws.cell(3, c).fill.start_color.value for c in range(1, 5)] == ["00F2F2F2"] * 4
    assert [ws.cell(3, c).font.bold for c in range(1, 5)] == [True] * 4

    assert [ws.cell(4, c).fill.fill_type for c in range(1, 5)] == ["solid"] * 4
    assert [ws.cell(4, c).fill.start_color.value for c in range(1, 5)] == ["00F2F2F2"] * 4
    assert [ws.cell(4, c).font.bold for c in range(1, 5)] == [False] * 4

    # - table data by column
    assert [ws.cell(r, 1).value for r in range(5, 9)] == ["home", "work", "beach", "wonderland"]
    assert [ws.cell(r, 1).fill.fill_type for r in range(5, 9)] == [None] * 4
    assert [ws.cell(r, 1).font.bold for r in range(5, 9)] == [False] * 4

    assert [ws.cell(r, 2).value for r in range(5, 9)] == [0, 1, 2, "-"]
    assert [ws.cell(r, 2).fill.fill_type for r in range(5, 9)] == [None] * 4
    assert [ws.cell(r, 2).font.bold for r in range(5, 9)] == [False] * 4
    for r, d in zip(
        range(5, 8), pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00"])
    ):
        # workaround openpyxl bug: https://foss.heptapod.net/openpyxl/openpyxl/-/issues/1493
        # openpyxl adds a spurious microsecond to some datetimes.
        assert abs(ws.cell(r, 3).value - d) <= datetime.timedelta(microseconds=1)
        assert ws.cell(r, 3).fill.fill_type is None
        assert ws.cell(r, 3).font.bold is False
    assert ws.cell(8, 3).value == "-"
    assert [ws.cell(r, 4).value for r in range(5, 9)] == [1, 0, 1, 0]
    assert [ws.cell(r, 4).fill.fill_type for r in range(5, 9)] == [None] * 4
    assert [ws.cell(r, 4).font.bold for r in range(5, 9)] == [False] * 4

    # Second table is there as well
    assert ws["A10"].value == "**bar*"
    assert ws["A10"].fill.fill_type == "solid"
    assert ws["A10"].fill.start_color.value == "00D9D9D9"
    assert ws["A10"].font.color.value == "001F4E78"
    assert ws["A10"].font.bold is True

    assert ws["A11"].value == "all"
    assert ws["A11"].fill.fill_type == "solid"
    assert ws["A11"].fill.start_color.value == "00D9D9D9"
    assert ws["A11"].font.color.value == "00808080"
    assert ws["A11"].font.bold is True

    # column headers (transposed)
    assert [ws.cell(r, 1).value for r in range(12, 14)] == ["digit", "spelling"]
    assert [ws.cell(r, 1).fill.fill_type for r in range(12, 14)] == ["solid"] * 2
    assert [ws.cell(r, 1).fill.start_color.value for r in range(12, 14)] == ["00F2F2F2"] * 2
    assert [ws.cell(r, 1).font.bold for r in range(12, 14)] == [True] * 2

    assert [ws.cell(r, 2).value for r in range(12, 14)] == ["-", "text"]
    assert [ws.cell(r, 2).fill.fill_type for r in range(12, 14)] == ["solid"] * 2
    assert [ws.cell(r, 2).fill.start_color.value for r in range(12, 14)] == ["00F2F2F2"] * 2
    assert [ws.cell(r, 2).font.bold for r in range(12, 14)] == [False] * 2

    # column values (transposed)
    assert [ws.cell(12, c).value for c in range(3, 6)] == [1, 6, 42]
    assert [ws.cell(12, c).fill.fill_type for c in range(3, 6)] == [None] * 3
    assert [ws.cell(12, c).font.bold for c in range(3, 6)] == [False] * 3

    assert [ws.cell(13, c).value for c in range(3, 6)] == ["one", "six", "forty-two"]
    assert [ws.cell(13, c).fill.fill_type for c in range(3, 6)] == [None] * 3
    assert [ws.cell(13, c).font.bold for c in range(3, 6)] == [False] * 3

    # Third table is there as well
    assert ws["A15"].value == "**bas*"
    assert ws["A15"].fill.fill_type == "solid"
    assert ws["A15"].fill.start_color.value == "00D9D9D9"
    assert ws["A15"].font.color.value == "001F4E78"
    assert ws["A15"].font.bold is True

    assert ws["A16"].value == "all"
    assert ws["A16"].fill.fill_type == "solid"
    assert ws["A16"].fill.start_color.value == "00D9D9D9"
    assert ws["A16"].font.color.value == "00808080"
    assert ws["A16"].font.bold is True

    # Teardown
    out_path.unlink()


def test_write_excel__custom_style(tmp_path):
    # Make a table
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    nc = len(t.column_names)
    nr = len(t.df)

    # Make a style specification as a JSON-like data structure
    style_spec = {
        "table_name": {
            "font": {
                "name": "Times New Roman",
                "size": 24,
                "color": "FF0000",   # RGB hex color code
                "bold": True,
            },
            "fill": {
                "color": "00AAAAAA",  # leading 'aa' transparency values are accepted (but unused)
            },
        },
        "destinations": {
            "font": {
                "italic": True,
                "color": "0000FF",
            },
            "fill": {
                "color": "888888",
            },
        },
        "column_names": {
            "font": {
                "color": "444400",
                "bold": True,
            },
            "fill": {
                "color": "777777",
            },
        },
        "units": {
            "font": {
                "color": "440044",
            },  # --------------------- fill unspecified, leave untouched
        },
        "values": {
            "alignment": {
                "horizontal": "left",
            },
            "fill": {
                "color": "EEEEEE",
            },  # --------------------- font unspecified, leave untouched
        },
    }

    # Write tables to workbook, save, and re-load
    out_path = tmp_path / "foo_custom_style.xlsx"
    write_excel([t], out_path, styles=style_spec)
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # Check table formatting
    # table name
    assert ws["A1"].fill.fill_type == "solid"
    assert ws["A1"].fill.start_color.value == "00AAAAAA"
    assert ws["A1"].font.size == 24
    assert ws["A1"].font.color.value == "00FF0000"
    assert ws["A1"].font.name == "Times New Roman"
    assert ws["A1"].font.bold is True

    # destinations
    assert ws["A2"].fill.fill_type == "solid"
    assert ws["A2"].fill.start_color.value == "00888888"
    assert ws["A2"].font.color.value == "000000FF"
    assert ws["A2"].font.bold is False
    assert ws["A2"].font.italic is True

    # column names
    assert [ws.cell(3, c).fill.fill_type for c in range(1, nc+1)] == ["solid"] * nc
    assert [ws.cell(3, c).fill.start_color.value for c in range(1, nc+1)] == ["00777777"] * nc
    assert [ws.cell(3, c).font.color.value for c in range(1, nc+1)] == ["00444400"] * nc
    assert [ws.cell(3, c).font.bold for c in range(1, nc+1)] == [True] * nc

    # column units
    assert [ws.cell(4, c).fill.fill_type for c in range(1, nc+1)] == [None] * nc  # left as default
    assert [ws.cell(4, c).font.color.value for c in range(1, nc+1)] == ["00440044"] * nc
    assert [ws.cell(4, c).font.bold for c in range(1, nc+1)] == [False] * nc
    assert [ws.cell(4, c).alignment.horizontal for c in range(1, nc+1)] == [None] * nc

    # values
    assert [[ws.cell(4 + r, c).fill.fill_type for c in range(1, nc + 1)] for r in range(1, nr + 1)] == [["solid"] * nc] * nr
    assert [[ws.cell(4 + r, c).fill.start_color.value for c in range(1, nc + 1)] for r in range(1, nr + 1)] == [["00EEEEEE"] * nc] * nr
    assert [[ws.cell(4 + r, c).alignment.horizontal for c in range(1, nc + 1)] for r in range(1, nr + 1)] == [["left"] * nc] * nr


@pytest.mark.parametrize(
    "err_msg_match,style_spec",
    [
        ("Invalid.*table_name", {"table_name": {"font": {"size": "NOT_A_NUMBER!!!"}}}),
        ("Invalid.*table_name", {"table_name": {"font": {"color": "NOT_A_COLOR!!!"}}}),
        ("Invalid.*destinations", {"destinations": {"fill": {"color": "NOT_A_COLOR!!!"}}}),
        ("Invalid.*values", {"values": {"alignment": {"horizontal": "FOOBAR!!!"}}}),
    ]
)
def test_write_excel__raises_error_on_invalid_style_spec(tmp_path, err_msg_match, style_spec):
    # Make a table
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")

    # Invalid font size
    with pytest.raises(ValueError, match=err_msg_match):
        write_excel([t], tmp_path / "foo_invalid_style.xlsx", styles=style_spec)


def test_write_excel__transposed_table_units_and_values_are_centered_by_default(tmp_path):
    # Make a table
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.metadata.transposed = True
    nc = len(t.column_names)
    nr = len(t.df)

    # DEFAULT ALIGNMENT: CENTER
    # Write tables to workbook with default styles, save, and re-load
    out_path = tmp_path / "foo_custom_style.xlsx"
    write_excel([t], out_path, styles=True)  # <<< Default style
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # column units and values are centered
    assert [ws.cell(2 + c, 2).alignment.horizontal for c in range(1, nc+1)] == ["center"] * nc
    assert [[ws.cell(2 + c, 2 + r).alignment.horizontal for c in range(1, nc + 1)] for r in range(1, nr + 1)] == [["center"] * nc] * nr

    # DEFAULT ALIGNMENT NOT APPLIED when custom alignment is specified
    # Write tables to workbook with custom alignment styles, save, and re-load
    out_path = tmp_path / "foo_custom_style.xlsx"
    left = {"alignment": {"horizontal": "left"}}
    write_excel([t], out_path, styles={"units": left, "values": left})  # << Custom alignment
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # column units and values are not centered
    assert [ws.cell(2 + c, 2).alignment.horizontal for c in range(1, nc+1)] == ["left"] * nc
    assert [[ws.cell(2 + c, 2 + r).alignment.horizontal for c in range(1, nc + 1)] for r in range(1, nr + 1)] == [["left"] * nc] * nr


def test_write_excel__sep_lines(tmp_path):
    # Make a couple of tables
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    # This one is transposed
    t2 = Table(name="bar")
    t2.add_column("digit", [1, 6, 42], "-")
    t2.add_column("spelling", ["one", "six", "forty-two"], "text")
    t2.metadata.transposed = True

    # This one is also transposed
    t3 = Table(name="bas")
    t3.add_column("digit", [1, 6, 42], "-")
    t3.add_column("spelling", ["one", "six", "forty-two"], "text")
    t3.metadata.transposed = True

    # Write tables to workbook, save, and re-load
    out_path = tmp_path / "foo.xlsx"
    write_excel([t, t2, t3], out_path, sep_lines=2)
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # Tables start on the expected rows
    assert ws["A1"].value == "**foo"
    assert ws["A11"].value == "**bar*"
    assert ws["A17"].value == "**bas*"


def test_read_write_excel__round_trip_with_styles(tmp_path):
    """Round-trip reading and writing and re-reading preserves tables"""
    from pdtable import TableBundle, read_excel
    bundle = TableBundle(read_excel("pdtable/test/io/input/foo.xlsx"))
    out_path = tmp_path / "foo_styled.xlsx"
    # Doesn't crash on write
    write_excel(bundle, out_path, styles=True)
    # Re-read bundle is same as first one
    bundle2 = TableBundle(read_excel(out_path))
    for t, t2 in zip(bundle, bundle2):
        assert t.equals(t2)



# flake8: noqa

from .csv import read_csv, write_csv
from .excel import read_excel, write_excel
from .json import table_to_json_data, json_data_to_table
from .parsers.fixer import ParseFixer
from .parsers.blocks import parse_blocks
from . import load


import pytest
from textwrap import dedent
from pathlib import Path
import re

import pdtable
from pdtable.store import BlockType
from pdtable.io.load import load_files, make_location_trees, FileReader
from pdtable.table_origin import InputError


def test_file_reader():
    r = FileReader(sheet_name_pattern=re.compile(".*"))

    assert set(r.supported_extensions) == {"csv", "xlsx"}
    assert r.supported_filename_pattern.pattern == r".*\.(csv|xlsx)$"


@pytest.fixture
def input_folder() -> Path:
    return Path(__file__).parent / "input"


def test_include(input_folder):
    root_folder = input_folder / "with_include"
    res = {
        b.name: b
        for t, b in load_files(
            ["/"], root_folder=root_folder, csv_sep=";", file_name_start_pattern="(input|setup)_"
        )
        if t == BlockType.TABLE
    }

    assert set(res.keys()) == {"bar_table", "bar_abs_table"}
    input_location = res["bar_table"].metadata.origin.input_location

    spec_ref = list(
        zip(
            str(input_location.file.load_specification).split(";"),
            [
                """included as "bar.csv" from """,
                """included as "input_foo.csv" from """,
                """included as "/" from <root>""",
            ],
        )
    )
    spec_ref_ok = [s.startswith(ref) for s, ref in spec_ref]
    assert all(spec_ref_ok)

    location_trees = make_location_trees(res.values())
    tree_as_str = "\n".join(str(n) for n in location_trees)
    reference_str = dedent(
        f"""
    <root_folder: {root_folder}>
      Row 0 of 'input_foo.csv'
        bar_abs.csv
          **bar_abs_table
        bar.csv
          **bar_table
    """
    ).strip()
    assert tree_as_str == reference_str



def test_include_without_root(input_folder):
    root_folder = input_folder / "with_include"
    blocks = list(load_files([root_folder/"foo_relative.csv"], csv_sep=";"))

    res = {b.name: b for t, b in blocks if t == BlockType.TABLE}

    assert set(res.keys()) == {"bar_table", "bar_abs_table"}

def test_excel_all_files(input_folder):
    blocks = load_files(
        ["/"],
        root_folder=input_folder,
        # negative lookahead to avoid excel temp files...
        file_name_pattern=re.compile(r"(?!~\$).*\.xlsx$"),
    )
    res = [b for t, b in blocks if t == BlockType.TABLE]

    assert len(res) == 9


def test_issue_raised(input_folder):
    # error because only absolute path is accepted relative to root folder
    with pytest.raises(pdtable.table_origin.InputError):
        list(load_files(["somefile.xlsx"], root_folder=input_folder))


def test_excel_sheets(input_folder):
    blocks = list(
        load_files(
            ["file:/multipage.xlsx"],
            root_folder=input_folder,
            sheet_name_pattern=re.compile(r"^input.*"),
        )
    )
    res = {b.name for t, b in blocks if t == BlockType.TABLE}
    assert "setup_table" not in res
    assert res.issuperset({"places_to_go", "spelling_numbers"})

    blocks = list(
        load_files(
            ["file:/multipage.xlsx"],
            root_folder=input_folder,
            sheet_name_pattern=re.compile(r"^(input|setup).*", re.IGNORECASE),
        )
    )
    res = {b.name for t, b in blocks if t == BlockType.TABLE}
    assert res.issuperset({"places_to_go", "setup_table", "spelling_numbers"})


def test_include_loop(input_folder: Path):
    f = input_folder / "with_loop_include/load_include_loop.csv"
    with pytest.raises(InputError, match="Load location included multiple times"):
        blocks = list(load_files([f.absolute()]))





"""Interface to read/write Tables from/to an Excel workbook.

The only Excel I/O engine supported right now is 'openpyxl', but this module can
be extended to support others readers such as 'xlrd' and writers such as 'xlsxwriter'.

openpyxl (and eventually other engines) are not required at install time; only when the functions
requiring them (read_excel() or write_excel()) are called for the first time.

"""
import os
from os import PathLike
from pathlib import Path
from typing import Union, Callable, Iterable, BinaryIO, Dict, Optional
import re
import warnings
import logging

from .parsers.blocks import parse_blocks
from .parsers.fixer import ParseFixer
from .. import BlockType, Table
from ..store import BlockIterator
from ..table_origin import FilesystemLocationFile, InputIssueTracker, LocationFile, NullLocationFile

logger = logging.getLogger(__name__)


def read_excel(
    source: Union[str, PathLike, Path],
    *,
    origin: str = None,
    location_file: LocationFile = None,
    sheet_name_pattern: Optional[re.Pattern] = None,
    fixer: ParseFixer = None,
    to: str = "pdtable",
    filter: Callable[[BlockType, str], bool] = None,
    issue_tracker: InputIssueTracker = None,
) -> BlockIterator:
    """Reads StarTable blocks from an Excel workbook.
    # TODO copy most of read_csv() docstring over

    Reads StarTable blocks from an Excel workbook file at the specified path.
    Yields them one at a time as a tuple: (block type, block content)

    Args:
        source:
            Path of workbook to read.

        origin:
            Optional; File origin description/file name as str. May be shadowed by `location_file`.

        location_file:
            Optional; Origin of file as a LocationFile object.
            `location_sheet` takes precedence over `origin`. For file input default
            if input path with `origin` as optional input specification, for stream input
            default is a null context with `origin` as description.
        sheet_name_pattern:
            Optional[re.Pattern] = None;
            If specified, only sheets with name matching pattern will be loaded.
            Matching is done with ``match``, i.e. must match from start of string.


    Yields:
        Tuples of the form (block type, block content)
    """

    source_is_stream = hasattr(source, "read")
    if not source_is_stream:
        source = Path(source)

    # resolve location
    if location_file is None:
        if not source_is_stream:
            location_file = FilesystemLocationFile(local_path=source, load_specification=origin)
        else:
            location_file = NullLocationFile()
    elif origin is not None:
        warnings.warn(
            f"Input 'origin': {origin} is shadowed by input 'location_file': {location_file}."
        )

    try:
        from ._excel_openpyxl import read_sheets
    except ImportError as err:
        raise ImportError(
            "Unable to find a usable Excel engine. "
            "Tried using: 'openpyxl'.\n"
            "Please install openpyxl for Excel I/O support."
        ) from err

    def name_matches(name) -> bool:
        if sheet_name_pattern is None:
            return True
        return sheet_name_pattern.match(name) is not None

    for name, row_cell_iter in read_sheets(source):
        if not name_matches(name):
            logger.debug(f"Skipping sheet '{name}'")
            continue
        location_sheet = location_file.make_location_sheet(name)
        yield from parse_blocks(
            row_cell_iter,
            location_sheet=location_sheet,
            fixer=fixer,
            to=to,
            filter=filter,
            issue_tracker=issue_tracker,
        )


def write_excel(
    tables: Union[Table, Iterable[Table], Dict[str, Table], Dict[str, Iterable[Table]]],
    to: Union[str, os.PathLike, Path, BinaryIO],
    na_rep: str = "-",
    sep_lines: int = 1,
    styles: Union[bool, Dict] = False,
):
    """Writes one or more tables to an Excel workbook.

    Writes table blocks to an Excel workbook file.
    Values are formatted to comply with the StarTable standard where necessary and possible.

    This is a thin wrapper around parse_blocks(). The only thing it does is to present the contents
    of an Excel workbook as a Iterable of cell rows, where each row is a sequence of values.

    Args:
        tables:
            Table(s) to write.
            * If a single Table or an iterable of Tables, writes to one sheet with default name.
            * If a dict of {sheet_name: Table} or {sheet_name: Iterable[Table]}, writes tables to
              sheets with specified names.

        to:
            File path or binary stream to which to write.
            If a file path, then this file gets created/overwritten and then closed after writing.
            If a stream, then it is left open after writing; the caller is responsible for managing
            the stream.

        na_rep:
            Optional; String representation of missing values (NaN, None, NaT).
            If overriding the default '-', it is recommended to use another value compliant with
            the StarTable standard.

        sep_lines:
            Optional; Number of blank separator lines between tables.
            Default is 1.

        styles:
            Optional. Determines whether styles are applied to table blocks in the output workbook.
            * If bool(styles) is False (default), no styles are applied.
            * If True, default pdtable styles are applied (neutral shades of grey and dark blue).
            * Custom styles can be specified by passing a JSON-like structure of dictionaries.
              The top-level keys represent the parts of a table block and are any number of:
              {'table_name', 'destinations', 'column_names', 'units', 'values'}
              For each given table part, the value is a dictionary with any number of the following
              style element keys: {'font', 'fill', 'alignment'}
              The value for each of these keys is in turn a dictionary of valid {property: value}
              pairs.
              Example:
                {
                    'table_name': {
                        'font': {'color': '1F4E78', 'bold': True,},
                        'fill': {'color': 'D9D9D9',},  # RGB color code
                    },
                    'destinations': {
                        'font': {'color': '808080', 'italic': True,},
                        'fill': {'color': 'D9D9D9',},
                    },
                    'column_names': {
                        'fill': {'color': 'F2F2F2',},
                        'font': {'bold': True,},
                    },
                    'units': {'fill': {'color': 'F2F2F2',}},
                    'values': {'alignment': {'horizontal': 'center'}},
                }

             Colors are given as RGB hex codes of the form 'RRGGBB' e.g. 'FF0000' is red.
             Leading transparency digits are accepted but unused e.g. '42FF0000' will still be red.

             If a table part key is omitted, then no style is applied to that table part.
             Similarly, if a style element key is omitted, then no style is explicitly applied to
             that style element.
             In such cases, the default style is determined by the Excel writer engine (the default
             engine is openpyxl).
    """
    try:
        from ._excel_openpyxl import write_excel_openpyxl as write_excel_func

    except ImportError as err:
        raise ImportError(
            "Unable to find a usable spreadsheet engine. "
            "Tried using: 'openpyxl'.\n"
            "Please install openpyxl for Excel I/O support."
        ) from err

    write_excel_func(tables, to, na_rep, styles, sep_lines)



"""
The purpose of the `table_origin` module is to provide an object model of the origin of a table

The `table_origin` module defines the data structure for achieving this. The process of building
the object model is performed by the load system. The object model is attached to `Table`-objects
as a `TableOrigin` instance in the `.origin` field of `.metadata`.
"""
from __future__ import annotations

from typing import (
    Iterator,
    NamedTuple,
    Iterable,
)
from typing_extensions import Protocol  # Protocol is not available in python 3.7
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from pathlib import Path
import logging
import sys
import os
import subprocess
import html
import base64
import datetime


class LoadItem(NamedTuple):
    specification: str
    source: None | LoadLocation

    @property
    def source_identifier(self) -> str:
        return "<root>" if self.source is None else self.source.load_identifier

    def load_history(self) -> Iterator["LoadItem"]:
        """
        Return the load tree leading up to this load item

        Typical use:
        ```
        '\n'.join(f'included as "{li.specification} from "{li.source_identifier}"'
                  for li in load_specification.load_history())
        ```

        Example text representation
        ```
        included as "doreco:input_foo_DOR12345" from "/mp/input_include.csv@2021-01-02T1233" row 27
        included as "input_include.csv" from "/mp"
        included as "mp/" from "/input_all.csv@123123123" row 12
        included as "input_all.csv" from "/"
        included as "/" from "<root>"
        ```
        """
        yield self
        if self.source:
            yield from self.source.load_specification.load_history()

    def __str__(self) -> str:
        return ";".join(
            f'included as "{li.specification}" from '
            f'{li.source.interactive_identifier if li.source else "<root>"}"'
            for li in self.load_history()
        )


def interactive_open_uri(uri):
    """
    A cross-platform functionality for launching tool by URI
    """
    if sys.platform == "win32":
        os.startfile(uri)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, uri])


class LoadLocation(Protocol):
    @property
    @abstractmethod
    def local_folder_path(self) -> None | Path:
        """
        Used for resolving relative imports
        """
        pass

    @property
    @abstractmethod
    def load_specification(self) -> LoadItem:
        """
        The specification passed to Loader instance as LoadItem to retrieve this file

        The specification may not uniquely identify the file, as i may include partial
        record specifications (e.g. "use latest") resolved at load time.
        """
        pass

    @property
    @abstractmethod
    def load_identifier(self) -> str:
        """
        Unique identifier of loaded item

        Must be unique to allow import loop detection and input caching.
        For local files, this could be absolute path and modification time.
        """
        pass

    @property
    def interactive_identifier(self) -> str:
        pass

    @abstractmethod
    def interactive_open(self, read_only: bool = False):
        pass

    @abstractmethod
    def interactive_uri(self, read_only: bool = False) -> str:
        pass


class LocationFile(LoadLocation):
    """
    Represents a traceable load entity

    The load entity could be a file, a blob, an http response. The LocationFile instance
    should hold enough information to uniquely identify (and if possible allow recreation)
    of the block stream resulting of loading this entity.
    """

    @property
    @abstractmethod
    def local_path(self) -> None | Path:
        pass

    @abstractmethod
    def interactive_uri(
        self, read_only=True, sheet: None | str = None, row: None | int = None
    ) -> None | str:
        pass

    @property
    def local_folder_path(self) -> None | Path:
        return None if self.local_path is None else self.local_path.parent

    def interactive_open(self, read_only=True, sheet: None | str = None, row: None | int = None):
        uri = self.interactive_uri(sheet, row, read_only)
        interactive_open_uri(uri)

    def get_interactive_identifier(self, sheet: None | str = None, row: None | int = None) -> str:
        """
        Interactive identifier for location

        Defaults to load identifier.
        May be replaced with replacement better suited for interactive use.
        """
        s_loc = "" if sheet is None else f" Sheet '{sheet}'"
        r_loc = "" if row is None else f" Row {row}"

        return self.load_identifier + s_loc + r_loc

    @property
    def interactive_identifier(self) -> str:
        return self.get_interactive_identifier()

    def get_local_path(self) -> Path:
        """
        Always return path to local instance

        Will write contents to local storage if not available.
        """
        if self.local_path is not None:
            return self.local_path
        raise NotImplementedError("Automatic download not implemented")

    def make_location_sheet(self, sheet_name: None | str = None):
        return LocationSheet(file=self, sheet_name=sheet_name)

    def __str__(self) -> str:
        return self.interactive_identifier

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={repr(self.load_identifier)}, "
            f"spec={repr(self.load_specification)})"
        )


def _random_id() -> str:
    return base64.b32encode(os.urandom(5 * 2)).decode("ascii")


class NullLocationFile(LocationFile):
    """
    Null-implementation of LocationFile
    """

    def __init__(self, description: None | str = None, id: None | str = None):
        if description is None:
            description = "Unknown"
        if id is None:
            id = f"{description}-{_random_id()}"
        self._spec = LoadItem(description, source=None)
        self._load_identifier = id

    @property
    def load_specification(self) -> LoadItem:
        return self._spec

    @property
    def load_identifier(self) -> str:
        return self._load_identifier

    @property
    def local_path(self) -> None | Path:
        return None

    def interactive_uri(
        self, sheet: None | str = None, row: None | int = None, read_only=True
    ) -> None | str:
        return None


class FilesystemLocationFile(LocationFile):
    """
    Args:
        root_folder: If specified, interactive_identifier will be given relative to this
    """

    def __init__(
        self,
        local_path: Path,
        load_specification: None | LoadItem = None,
        root_folder: None | Path = None,
        stat_result=None,
    ):
        self._local_path = local_path
        self._load_specification = load_specification or LoadItem(
            specification=str(local_path), source=None
        )
        self._stat_result = stat_result
        self.root_folder = root_folder

    @property
    def local_path(self):
        return self._local_path

    @property
    def load_specification(self) -> LoadItem:
        return self._load_specification

    def get_stat_result(self, cached=True):
        if (not cached) or self._stat_result is None:
            self._stat_result = self.local_path.stat()
        return self._stat_result

    def get_mtime(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.get_stat_result().st_mtime)

    @property
    def load_identifier(self) -> str:
        name_part = str(self.local_path.absolute())
        mtime = self.get_mtime()
        if mtime:
            return name_part + "@" + mtime.isoformat(timespec="seconds")
        return name_part

    @property
    def interactive_identifier(self) -> str:
        if self.root_folder is None:
            return str(self.local_path)
        return str(self.local_path.relative_to(self.root_folder))

    def get_interactive_identifier(self, sheet: None | str = None, row: None | int = None) -> str:
        if sheet is None:
            loc = f"Row {row}"
        else:
            loc = f"'{sheet}'!A{row}"
        return f"{loc} of '{self.interactive_identifier}'"

    def interactive_uri(
        self,
        sheet: None | str = None,
        row: None | int = None,
        read_only: None | bool = False,
    ) -> str:
        file_uri = self.local_path.as_uri()
        if sheet is None and row is None:
            return file_uri
        if sheet is None:
            sheet = "Sheet1"
        row_mark = f"!A{row}" if row is not None else ""
        return file_uri + f"#'{sheet}'{row_mark}"


@dataclass(frozen=True)  # to allow empty dict default
class LocationSheet:
    file: LocationFile
    sheet_name: None | str
    sheet_metadata: dict[str, str] = field(default_factory=dict)

    def make_location_block(self, row: int):
        return LocationBlock(sheet=self, row=row)


class LocationBlock(NamedTuple):
    """ """

    sheet: LocationSheet
    row: int

    @property
    def file(self) -> LocationFile:
        return self.sheet.file

    @property
    def sheet_name(self) -> None | str:
        return self.sheet.sheet_name

    @property
    def local_folder_path(self) -> None | Path:
        return self.file.local_folder_path

    @property
    def load_identifier(self) -> str:
        return f"{self.file.load_identifier}#'{self.sheet_name or 'Sheet1'}'!A{self.row}"

    @property
    def load_specification(self) -> LoadItem:
        return self.file.load_specification

    @property
    def interactive_identifier(self) -> str:
        return self.file.get_interactive_identifier(sheet=self.sheet_name, row=self.row)

    def interactive_uri(self, read_only: bool = False):
        return self.file.interactive_uri(sheet=self.sheet_name, row=self.row, read_only=read_only)

    def interactive_open(self, read_only: bool = False):
        interactive_open_uri(self.interactive_uri(read_only=read_only))

    def __str__(self) -> str:
        return f"{self.interactive_identifier};{self.file.load_specification}"

    def __repr__(self) -> str:
        sheet_spec = "" if self.sheet_name is None else f", sheet={self.sheet_name}"
        return f"LocationBlock(row={self.row}{sheet_spec}, file={repr(self.file)})"


@dataclass(frozen=True)
class TableOrigin:
    """
    A TableOrigin instance defines the source of a Table instance.

    The source may either be a loaded input table or an operation combining multiple
    parents (represented by TableOrigin instances) into a derived table, so that each
    ``TableOrigin``-instance is the root of a tree of ``TableOrigin`` instances.

    Each node in the tree is either a _leaf_, corresponding to a loaded input, in which
    case only ``input_location`` is defined, or a _branch_, corresponding to a derived
    table, in which case only ``parents`` and ``operation`` are defined.

    For an integrated representation of this information, an application should traverse
    the tree and provide a representation in the most convenient form. For an example,
    see ``table_origin_as_html`` or ``table_origin_as_str``.
    """

    input_location: None | LocationBlock = None
    parents: Iterable["TableOrigin"] = ()
    operation: None | str = None

    def _post_init_(self):
        if self.operation is None:
            # leaf
            if self.parents or self.input_location is None:
                raise ValueError("For TableOrigin leaf node, only input_location must be defined")
        else:
            # branch
            if self.input_location is not None or not self.parents:
                raise ValueError(
                    "For TableOrigin branch node, only operation and parents should be defined"
                )

    @property
    def is_leaf(self):
        return self.operation is None

    def get_input_ancestors(self) -> Iterator[LocationBlock]:
        """
        Return iterator over the input-location of all non-derived ancestors
        """
        if self.is_leaf:
            if self.input_location:
                yield self.input_location
            else:
                raise ValueError("Inconsistent state of TableOrigin")
        else:
            for p in self.parents:
                yield from p.get_input_ancestors()

    def __str__(self):
        return table_origin_as_str(self)

    def _repr_html_(self) -> str:
        return table_origin_as_html(self)


def table_origin_as_html(tt: TableOrigin):
    def visit(tt):
        return visit_branch(tt) if tt.input_location is None else visit_leaf(tt)

    def visit_leaf(tt):
        loc = tt.input_location
        return (
            f"""<a href="{loc.interactive_uri()}" class="input-table-origin">"""
            f"""{html.escape(loc.interactive_identifier)}</a>"""
        )

    def visit_branch(tt):
        return (
            f"""<div class="derived-table-origin"><span>{html.escape(tt.operation)}</span><ul>"""
            + "\n".join(f"<li>{visit(p)}</li>" for p in tt.parents)
            + """</ul></div>"""
        )

    return visit(tt)


def table_origin_as_str(tt: TableOrigin):
    buf: list[tuple[int, str]] = []

    def visit(tt, lev):
        return visit_branch(tt, lev) if tt.input_location is None else visit_leaf(tt, lev)

    def visit_leaf(tt, lev):
        buf.append((lev, tt.input_location.interactive_identifier))

    def visit_branch(tt, lev):
        buf.append((lev, f"Derived via {tt.operation} from:"))
        for p in tt.parents:
            visit(p, lev + 1)

    visit(tt, 0)
    return "\n".join("  " * lev + s for lev, s in buf)


@dataclass(frozen=True)
class InputIssue:
    issue: str | Exception
    load_item: None | LoadItem = None
    load_location: None | LoadLocation = None
    severity: int = logging.ERROR

    def __post_init__(self):
        # Set load_item according to load_location if applicable
        if self.load_item is None and self.load_location is not None:
            object.__setattr__(self, "load_item", self.load_location.load_specification)


class InputIssueTracker(ABC):
    """
    Protocol for tracking issues across inputs
    """

    @abstractmethod
    def add_issue(self, input_issue: InputIssue):
        pass

    def add_error(
        self,
        issue: str | Exception,
        load_item: None | LoadItem = None,
        load_location: None | LocationFile = None,
    ):
        """
        Add a error for critical input issue that should cause load to abort

        This is a convenience function that will create an `InputIssue` instance with
        severity "error" and pass this to the ``.add_issue`` method.
        """
        self.add_issue(
            InputIssue(
                load_item=load_item,
                load_location=load_location,
                issue=issue,
                severity=logging.ERROR,
            )
        )

    def add_warning(
        self,
        issue: str | Exception,
        load_item: None | LoadItem = None,
        load_location: None | LocationFile = None,
    ):
        """
        Add a warning about a non-critical input issue

        This is a convenience function that will create an `InputIssue` instance with
        severity "warning" and pass this to the ``.add_issue`` method.

        Examples include
        - additional columns compared to template
        - additional tables compared to template
        """
        self.add_issue(
            InputIssue(
                load_item=load_item,
                load_location=load_location,
                issue=issue,
                severity=logging.WARNING,
            )
        )

    @property
    @abstractmethod
    def issues(self) -> Iterable[InputIssue]:
        pass

    @property
    def is_ok(self) -> bool:
        """
        True if no errors have been registered
        """
        return not any(m.severity >= logging.ERROR for m in self.issues)


class InputError(Exception):
    """
    An exception raised on irrecoverable error in the input processing

    This exception should now be caught within the pdtable framework.
    """

    pass


class NullInputIssueTracker(InputIssueTracker):
    """
    Log the issue and raise an Input Error immediately if severity is not below logging.ERROR

    Note that this tracker is stateless: `.is_ok()` will always return `True`, and `.issues()`
    will always return empty.
    """

    def add_issue(self, input_issue: InputIssue):
        logging.log(input_issue.severity, str(input_issue))
        if input_issue.severity >= logging.ERROR:
            raise InputError(input_issue)

    @property
    def is_ok(self):
        return True

    @property
    def issues(self):
        return ()




"""Parsers to convert uncontrolled cell grids into representations of StarTable blocks.

parse_blocks() emits a stream of blocks objects.
This in principle allows early abort of reads as well as generic postprocessing (
as discussed in store-module docstring).

parse_blocks() switches between different parsers depending on the StarTable block type:
- Metadata
- Directive
- Table
- Template (not yet implemented)

For each of these:

- The intended input is a two-dimensional "cell grid" i.e. a sequence of rows, with each row
  being a sequence of values (where a "sequence" is usually a list, but can also be e.g. a tuple).
  Rows need not be of equal length; only the relevant portion of a cell grid will be parsed
  depending on the relevant block type.

- The output is a representation of the StarTable block, either as:
  - A pdtable-style block object e.g. Table
  - A JSON-like data structure ready for serialization via e.g. json.dump()  (only implemented for
    table blocks at this stage); or
  - The original, raw cell grid, in case the user wants to do some low-level processing.

"""
from abc import abstractmethod
import itertools
import re
from typing import Sequence, Optional, Tuple, Any, Iterable, List, Union, Dict
from collections import defaultdict
import pandas as pd
import warnings

from pdtable import BlockType, BlockIterator
from pdtable import Table
from pdtable.io._json import to_json_serializable, JsonData, JsonDataPrecursor
from pdtable.table_origin import (
    LocationSheet,
    NullLocationFile,
    TableOrigin,
    InputIssue,
    InputIssueTracker,
    NullInputIssueTracker,
)
from .columns import parse_column
from .fixer import ParseFixer
from ... import frame
from ...auxiliary import MetadataBlock, Directive
from ...table_metadata import TableMetadata

# Typing alias: 2D grid of cells with rows and cols. Intended indexing: cell_grid[row][col]
CellGrid = Sequence[Sequence]


def make_metadata_block(cells: CellGrid, origin: Optional[str] = None, **_) -> MetadataBlock:
    mb = MetadataBlock(origin)
    for row in cells:
        if len(row) > 1 and row[0] is not None:
            key_field = row[0].strip()
            if len(key_field) > 0 and key_field[-1] == ":":
                mb[key_field[:-1]] = row[1].strip()
    return mb


def make_directive(cells: CellGrid, origin: Optional[str] = None, **_) -> Directive:
    name = cells[0][0][3:]
    directive_lines = [row[0] for row in cells[1:]]
    return Directive(name, directive_lines, origin)


def make_fixer(origin, fixer=None, **kwargs):
    """ Determine if user has supplied custom fixer
        Else return default ParseFixer() instance.
    """
    if fixer is not None:
        if type(fixer) is type:
            # It's a class, not an instance. Make an instance here.
            fixer = fixer()
    else:
        fixer = ParseFixer()
    assert fixer is not None
    fixer.origin = origin
    # fixer.reset_fixes()
    return fixer


def parse_column_names(column_names_raw: Sequence[Union[str, None]]) -> List[str]:
    """Parses column names from the sequence read from file

    Rejects everything after first blank cell, since there can be comments there.
    Strips column names. 
    """
    return [
        c.strip() for c in itertools.takewhile(lambda x: not _is_cell_blank(x), column_names_raw)
    ]


def make_table_json_precursor(cells: CellGrid, origin, fixer:ParseFixer) -> Tuple[JsonDataPrecursor, bool]:
    """Parses cell grid into a JSON-like data structure but with some non-JSON-native values

    Parses cell grid to a JSON-like data structure of nested "objects" (dict), "arrays" (list),
    and values, including values with types that map 1:1 to JSON-native types, as well as some
    value types that don't directly map to JSON types.

    This JSON data "precursor" can then be sent for further processing:
    - Parsing to pdtable-style Table block object
    - Conversion to a "pure" JSON data object in which all values are of JSON-native types.

    Also returns a bool "transposed" flag.
    """

    table_name: str = cells[0][0][2:]
    transposed = table_name.endswith("*")
    if transposed:
        # Chop off the transpose decorator from the name
        table_name = table_name[:-1]
    fixer.table_name = table_name

    # internally hold destinations as json-compatible dict
    destinations = {dest: None for dest in cells[1][0].strip().split(" ")}

    if transposed:
        # Column names are in lines' first cell
        column_names = parse_column_names([line[0] for line in cells[2:]])
    else:
        # Column names are on line 2 (zero-based)
        column_names = parse_column_names(cells[2])
    column_names = _fix_duplicate_column_names(column_names, fixer)

    n_col = len(column_names)
    if transposed:
        units = [line[1] for line in cells[2 : 2 + n_col]]
    else:
        units = cells[3][:n_col]
    units = [unit.strip() for unit in units]

    if transposed:
        data_lines = [line[2:] for line in cells[2 : 2 + n_col]]
        len_longest_line = max(len(line) for line in data_lines)

        # Find last non-blank data row
        n_row = 0
        for i_row in range(len_longest_line):
            # Does this row have non-blank cells?
            for line in data_lines:
                if len(line) >= i_row + 1 and not _is_cell_blank(line[i_row]):
                    # Found a non-blank cell. This row is legit. Go check next row.
                    n_row = i_row + 1
                    break
            else:
                # No non-blank cells found on this row. This row is blank. Go no further.
                break

        # Collate data rows
        data_rows = zip(
            *(
                line[:n_row]  # trim empty cells off of long lines
                if len(line) >= n_row
                else line + [None] * (n_row - len(line))  # pad short lines with empty cells
                for line in data_lines
            )
        )
    else:
        data_rows = [line[:n_col] for line in cells[4:]]
    data_rows = [list(row) for row in data_rows]

    # ensure all data columns are populated
    for i_row, row in enumerate(data_rows):
        if len(row) < n_col:
            fix_row = fixer.fix_missing_rows_in_column_data(
                row=i_row, row_data=row, num_columns=n_col
            )
            data_rows[i_row] = fix_row

    # build dictionary of columns iteratively to allow meaningful error messages
    columns = dict(zip(column_names, [[]] * len(column_names)))
    for name, unit, values in zip(column_names, units, zip(*data_rows)):
        try:
            fixer.column_name = name
            columns[name] = parse_column(unit, values, fixer)
        except ValueError as e:
            raise ValueError(
                f"Unable to parse value in column '{name}' of table '{table_name}' as '{unit}'"
            ) from e

    fixer.report()

    return (
        {
            "name": table_name,
            "columns": columns,
            "units": units,
            "destinations": destinations,
            "origin": origin,
        },
        transposed,
    )


def _make_table(cells: CellGrid, origin, fixer) -> Table:
    """Parses cell grid into a pdtable-style Table block object."""
    json_precursor, transposed = make_table_json_precursor(
        cells, origin=str(origin.input_location), fixer=fixer,
    )
    return Table(
        frame.make_table_dataframe(
            pd.DataFrame(json_precursor["columns"]),
            units=json_precursor["units"],
            table_metadata=TableMetadata(
                name=json_precursor["name"],
                destinations=set(json_precursor["destinations"].keys()),
                origin=origin,
                transposed=transposed,
            ),
        )
    )


def make_table(cells: CellGrid, origin: Optional[TableOrigin]=None, **kwargs) -> Table:
    """Parses cell grid into a pdtable-style Table block object."""
    fixer=make_fixer(origin=origin, **kwargs)
    if origin is None:
        origin = TableOrigin()
    elif isinstance(origin, str):
        warnings.warn("Passing origin as str is deprecated", DeprecationWarning, stacklevel=2)
        origin = TableOrigin(NullLocationFile(origin).make_location_sheet().make_location_block(0))
    return _make_table(cells, origin, fixer=fixer)


def make_table_json_data(cells: CellGrid, origin, fixer) -> JsonData:
    """Parses cell grid into a JSON-ready data structure."""
    impure_json, transposed = make_table_json_precursor(cells, origin=origin, fixer=fixer)
    # attach unit directly to individual column
    units = impure_json["units"]
    del impure_json["units"]  # replaced by "unit" field in columns
    del impure_json["origin"]  # not relevant for json_data
    columns = {}
    for cname, unit in zip(impure_json["columns"].keys(), units):
        columns[cname] = {"unit": unit, "values": impure_json["columns"][cname]}
    impure_json["columns"] = columns
    return to_json_serializable(impure_json)


def make_raw_cells(cells: CellGrid, origin, **kwargs) -> CellGrid:
    return cells


DEFAULT_HANDLERS = (
    (BlockType.METADATA, make_metadata_block),
    (BlockType.DIRECTIVE, make_directive),
    (BlockType.TABLE, _make_table),
)
_default_handlers = dict(DEFAULT_HANDLERS)

TABLE_HANDLERS = (
    ("pdtable", _make_table),
    ("jsondata", make_table_json_data),
    ("cellgrid", make_raw_cells),
)
_table_handlers = dict(TABLE_HANDLERS)

# Regex that matches valid block start markers
_re_block_marker = re.compile(
    r"^("  # Marker must start exactly at start of cell
    r"(?<!\*)(\*\*\*?)(?!\*)"  # **table or ***directive but not ****undefined
    r"|"
    r"((?<!:):{1,3}(?!:))[^:]*\s*$"  # :col, ::table, :::file but not ::::undefined, :ambiguous:
    r"|"
    r"([^:]+:)\s*$"  # metadata:  but not :ambiguous:
    r")"
)
# $1 = Valid block start marker
# $2 = **Table / ***Directive
# $3 = :Template
# $4 = Metadata:


def _apply_filter(block_type, filter, handler):
    if not block_type == BlockType.TABLE:
        return lambda cellgrid, *args, **kwargs: handler(cellgrid, *args, **kwargs) if filter(
            block_type, ""
        ) else None
    return (
        lambda cellgrid, *args, **kwargs: handler(cellgrid, *args, **kwargs)
        if filter(block_type, cellgrid[0][0][2:])
        else None
    )


def parse_blocks(
    cell_rows: Iterable[Sequence],
    location_sheet: LocationSheet = None,
    to: str = "pdtable",
    filter: Any = None,
    fixer: Any = None,
    issue_tracker: InputIssueTracker = None,
    origin: Optional[str] = None,
    **kwargs,
) -> BlockIterator:
    """Parses blocks from a single sheet as rows of cells.

    This is a legacy facade for `parse_blocks_stable`.

    Takes an iterable of cell rows and parses it into blocks.

    Args:
        cell_rows: 
            Iterable of cell rows, where each row is a sequence of cells.
        location_sheet: 
            A `LocationSheet` object describing the sheet being read.
        filter: 
            Optional. Will be called as (block type, name | ""). Block is dropped if false.
        to: 
            Optional. Generate Table of this type ("pdtable", "jsondata", "cellgrid")
    kwargs:
        fixer: Also a thing, but different.
    Yields:
        Blocks.
    """

    # Set up handlers
    # Legacy default is to emit unknown types as raw cells
    handlers = {
        bt: make_raw_cells for bt in BlockType
    }  
    handlers.update(DEFAULT_HANDLERS)
    try:
        handlers[BlockType.TABLE] = _table_handlers[to]
    except KeyError:
        raise ValueError(
            f"Unknown parsing output type; expected one of {list(_table_handlers.keys())}.", to
        )
    if filter:
        for k, base_handler in handlers.items():
            handlers[k] = _apply_filter(k, filter, base_handler)
    if origin and location_sheet:
        warnings.warn("Origin is shadowed by location.sheet.")

    if fixer is not None or kwargs:
        origin_str = location_sheet.file.load_identifier if location_sheet is not None else origin
        fixer = make_fixer(origin=origin_str, fixer=fixer, **kwargs)
        warnings.warn(
            "The fixer construct is deprecated and will be removed in future release. "
            "Please file an issue describing fixer use case at https://github.com/startable/pdtable/issues."
            f"Fixer was triggered by the following keyword args: fixer: {fixer}, [{kwargs}].",
            DeprecationWarning,
            stacklevel=2,
        )
    else:
        fixer = None

    yield from parse_blocks_stable(
        cell_rows,
        location_sheet=location_sheet,
        block_handlers=handlers,
        fixer=fixer,
        issue_tracker=issue_tracker,
    )


# Regex that matches valid block start markers
_re_block_marker = re.compile(
    r"^("  # Marker must start exactly at start of cell
    r"(?<!\*)(\*\*\*?)(?!\*)"  # **table or ***directive but not ****undefined
    r"|"
    r"((?<!:):{1,3}(?!:))[^:]*\s*$"  # :col, ::table, :::file but not ::::undefined, :ambiguous:
    r"|"
    r"([^:]+:)\s*$"  # metadata:  but not :ambiguous:
    r")"
)
# $1 = Valid block start marker
# $2 = **Table / ***Directive
# $3 = :Template
# $4 = Metadata:


def parse_blocks_stable(
    cell_rows: Iterable[Sequence],
    issue_tracker: InputIssueTracker = None,
    block_handlers: Dict[BlockType, Any] = None,
    location_sheet: LocationSheet = None,
    fixer: Any = None,
) -> BlockIterator:
    """
    Generate blocks (tables, metadata, directives,...) from cell-rows

    This is a utility function for use by format-specific readers. It will return
    a block generator based on a stream of cell rows.
    The actual blocks are built by the supplied block handlers. A reader may need to 
    modify the defaults to correctly handle format conversions.

    Block handlers:
    The supplied block_handlers dictionary is queried for a handler by block type.
    If no block handler is defined, block will be silently ignored.
    Otherwise handler is called with arguments (cell_grid, origin, fixer) and should return
    a block of the correct type. If `None` is returned the block is silently ignored.

    Args:
        issue_tracker:
            Read issues are handled by calls to the issue tracker. The default issue tracker
            will simply raise an exception.
        block_handlers:
            A dictionary mapping block types to block handlers as described above
        location_sheet:
            Describes input location
        fixer:
            The fixer construct is a legacy system which will be deprecated. Please raise
            issues on github for all use-cases.
    """
    if location_sheet is None:
        location_sheet = NullLocationFile().make_location_sheet()

    if issue_tracker is None:
        issue_tracker = NullInputIssueTracker()

    if block_handlers is None:
        block_handlers = dict(DEFAULT_HANDLERS)

    if fixer is None:
        fixer = make_fixer(origin=location_sheet.file.load_identifier)

    def block_output(block_type, cell_grid, row: int):
        """
        Emit cell_grid as given block_type
        """
        if not cell_grid:  # return on None or empty
            return
        handler = block_handlers.get(block_type, None)
        if handler is None:
            return
        origin = TableOrigin(input_location=location_sheet.make_location_block(row=row))

        fixer.reset_fixes()
        try:
            block = handler(cell_grid, origin=origin, fixer=fixer)
        except ValueError as e:
            issue_tracker.add_error(str(e), load_location=origin.input_location)

        if block is not None:
            yield block_type, block

    cell_grid = []
    state = BlockType.METADATA
    next_state = None
    this_block_1st_row = 0
    for row_number_0based, row in enumerate(cell_rows):
        if row is None or len(row) == 0 or _is_cell_blank(row[0]):
            if state != BlockType.BLANK:
                next_state = BlockType.BLANK
            else:
                continue
        elif isinstance(row[0], str):
            # possible token
            mm = _re_block_marker.match(row[0])
            if mm is None:  # TBC (PEP 572)
                cell_grid.append(row)
                continue

            if mm.group(1) is not None:
                if mm.group(1) == "**":
                    next_state = BlockType.TABLE
                elif mm.group(1) == "***":
                    next_state = BlockType.DIRECTIVE
                elif mm.group(4) is not None:
                    if state == BlockType.METADATA:
                        cell_grid.append(row)
                        continue
                    else:
                        next_state = BlockType.BLANK
                else:
                    next_state = BlockType.TEMPLATE_ROW
        else:
            # binary (excel &c.)
            cell_grid.append(row)
            continue

        if next_state is not None:
            # Current block has ended. Emit it.
            yield from block_output(state, cell_grid, this_block_1st_row)
            cell_grid = []
            state = next_state
            next_state = None
            this_block_1st_row = row_number_0based
            if state != BlockType.BLANK:
                cell_grid.append(row)
            elif len(row) > 0:
                #  emit non-empty lines, comments &c. as BLANK
                if len(row) == 1 and _is_cell_blank(row[0]):
                    continue
                cell_grid.append(row)

    yield from block_output(state, cell_grid, this_block_1st_row)


def _fix_duplicate_column_names(col_names_raw: Sequence[str], fixer: ParseFixer):
    """Finds duplicate column names and sends them to ParseFixer for fixing."""
    column_names = []
    names = {}
    for col, cname in enumerate(col_names_raw):
        if cname not in names and len(cname) > 0:
            names[cname] = 0
            column_names.append(cname)
        else:
            fixer.column_name = col
            if cname in names:
                cname = fixer.fix_duplicate_column_name(cname, input_columns=column_names)
            assert cname not in names
            names[cname] = 0
            column_names.append(cname)
    return column_names


def _is_cell_blank(cell):
    """Is this cell blank i.e. contains nothing or only whitespace"""
    return cell is None or (isinstance(cell, str) and not cell.strip())
